/***************************************************************************
This macro will allow the caller to do a batch update to the SPM database.

The following options must be set prior to running this macro and should
reference the metadata server being used by StM.  For example:
metaserver=seshp13
metaport=8561
metarepository=Foundation


Required General Parameters:

USER                    Userid of who is performing the update
PW                      Password for the user.  It is advisable to use
                          an encrypted value. To encrypt a password, use the following
                          SAS code: proc pwencode in='pw_to_encrypt'; run;
PROJECTNAME             Name of the project
TEMPLATENAME            Name of the template that contains the project
OUTPUTDIR               The directory BMF writes the output CSV files to.
                          This directory must be accessible from the midtier


Optional General Parameters:

ACTION                  Valid values are CREATE, MODIFY, or GET. Defaults to CREATE. Numeric
                          values can be used where 1 = CREATE 2 = MODIFY and 3 = GET.
SYNCHRONOUS             Run the macro synchronously or asynchronously.  If synchronously, then
                          the macro will wait for the BMF task to finish before allowing
                          subsequent SAS statements to run.  Valid values are YES, Y, or 1 to
                          run synchronously.  If the parameter specifies any other value or the
                          parameter is not specified then the default is to run asynchronously.


These parameters specify file input and have precedence over their analogous SAS data
set input parameters.  The specified files must be accessible from the midtier (i.e.
use share and \\machinename if this SAS job is running on a different machine than the
midtier).  All of these parameters are optional and are only used when the specific file
is needed for the batch operation being requested.

SCORECARD               Scorecard data file
ATTRIBUTE               Attribute data file
ELEMENT                 Element data file
CELL                    Cell data file
CELLFORMAT              Cell format data file
SETUP                   File for data associated with the template such as element types,
                          metric attributes, templates and attribute definitions
PROJECT                 Project data file
RANGE                   Range data file
DIAGRAMS                Diagram data file
FILTER                  Filter data file


These parameters specify SAS Data set input.  Generally specify either a file parameter or
a SAS data set parameter, but not both.  For example, specify either scorecard=<value>  or
scorecardds=<value> but not both.

SCORECARDDS             Scorecard SAS data set
ATTRIBUTEDS             Attribute SAS data set
ELEMENTDS               Element SAS data set
CELLDS                  Cell SAS data set
CELLFORMATDS            Cell format SAS data set
SETUPDS                 SAS data set for data associated with the template such as element
                          types, metric attributes, templates and attribute definitions
PROJECTDS               Project SAS data set
RANGEDS                 Range SAS data set
FILTERDS                Filter SAS data set


An alternative to specifying individual files or SAS data sets is to use the INPUTDIR,
INPUTDSLIB or OUTPUTDSLIB parameters:

INPUTDIR                Directory containing appropriately named files and directories to be
                          used as input files when the value for the ACTION parameter is CREATE
                          or MODIFY.
                          The INPUTDIR directory will be searched for a subdirectory named the same
                          as the value of the TEMPLATENAME parameter. If this subdirectory
                          contains a file named <PROJECTNAME>_setup.csv, it will be used
                          but only if the SETUP= parameter does not have a value.  Additionally,
                          if this subdirectory contains a subdirectory named the same as the
                          value of the PROJECTNAME= parameter, then the following files will
                          be used as input files.  These files are named <PROJECTNAME>_scorecard.csv,
                          <PROJECTNAME>_attribute.csv, <PROJECTNAME>_element.csv,
                          <PROJECTNAME>_cell.csv, <PROJECTNAME>_setup.csv,
                          <PROJECTNAME>_ranges.csv, <PROJECTNAME>_filter.csv.
                          These files are used only if their analogous parameter does not have
                          an explicit value.  For example, even in there is a file named
                          <PROJECTNAME>_scorecard.csv, it will not be used if there is an explicit
                          value for the SCORECARD parameter. The INPUTDIR parameter is only valid
                          when the ACTION parameter is CREATE or MODIFY and the PROJECTID
                          parameter is not used.

INPUTDSLIB              SAS libref containing appropriately named SAS data set used as
                          input files.  These data sets are named <projectname>_scorecard,
                          <projectname>_attribute, <projectname>_element, <projectname>_cell,
                          <projectname>_setup, <projectname>_ranges, <projectname>_filter.
                          These datasets are used only if their analogous parameter does not
                          have an explicit value.
                          This parameter is only valid when the ACTION parameter is
                          CREATE or MODIFY and the PROJECTID parameter is not used.

OUTPUTDSLIB             SAS data library that the output SAS datasets are stored.  They
                          will be named <projectname>_scorecard,  <projectname>_attribute,
                          <projectname>_element, <projectname>_cell, <projectname>_setup,
                          <projectname>_ranges, <projectname>_filter. This is only enabled
                          when using the synchronous event.

CONVERTEDDSDIR          Directory to hold the CSV files that have been converted from the
                          input SAS datasets. This parameter is required when one or more
                          input SAS data sets are supplied or the INPUTDSLIB parameter is used.
                          The directory path must be accessible from the midtier (i.e.
                          use share and \\machinename if this SAS job is running on a
                          different machine than the midtier).

Misc other optional parameters:

PROJECTID               GUID of the project.  This can be used instead of the
                          projectname and templatename parameters.  It is useful
                          because the GUID is locale independent.
BMFRCVAR                An existing macro variable that will hold a return code
                          from BMF. Any non-zero value indicates an error.
MIGRATE                 Tells BMF that this invocation is intended to migrate a Project.
                          Valid values are YES, Y, or 1 to migrate. If the parameter specifies
                          any other value or the parameter is not specified then the
                          default is to not migrate.  (Not implemented until V5.1)


These parameters are used for exporting to IMap or are not necessary for
typical BMF functions:

PRIORITY                Defaults to a valid value.
EVENTNAME               Defaults to the asynchronous event (SAS.Solutions.SpmBOBInterface).
                          If synchronous processing is desired, specify the
                          SAS.Solutions.SpmBOBInterfaceSynch event. Normally this parameter
                          is not needed because the SYNCHRONOUS parameter is the preferred
                          method of determining whether the macro runs synchronously.
INVALIDDIRCHARS         List of characters that may be in the project or template name but
                          cannot be used for directory names.  There is an extensive default
                          value for this parameter and the parameter is not normally needed.
DOMAIN                  Domain that the user logs in with.
ENTITYKEY               Required only if no user/pw is specified,  the entityKey for the
                          user's session.
SCORECARDID             GUID of the scorecard to be exported.
ELEMENTTYPEID           GUID of the element type selected at time of export.
DATE                    Date value, if set to CURRENT then it will use the current date.
OFFSET                  Offset value in time periods from specified date.
NUMPERIODS              Number of time periods to fetch back from the specified date.
                          Use '*' for all.
SORT                    Element sorting at time of export.
FILTERID                Filtering in place at time of exporting.
CULTUREID               GUID of the cultureid at time of exporting.
NAMEVAR                 Value of the Name column in the output data.
SCORECARDVAR            Value of the Scorecard column in the output data.
BADNUMBER               Value to use in case of a bad number.
OUTPUTTYPE              User can specify "DAV" as the output type.
RESULTTYPE
GUESSINGROWS            Used for CVS to SAS data set conversion.  Defaults to 32767
SENTBY                  Defaults to a descriptive value
****************************************************************************/

%macro spmbmf(
   action=create,
   attribute=,
   attributeds=,
   badnumber=,
   bmfrcvar=,
   cell=,
   cellds=,
   cellformat=,
   cellformatds=,
   converteddsdir=,
   cultureid=,
   date=,
   diagrams=,
   domain=,
   element=,
   elementds=,
   elementtypeid=,
   entitykey=x,
   eventName=,
   filter=,
   filterds=,
   filterid=,
   guessingrows=32767,
   inputdir=,
   inputdslib=,
   invalidDirChars=,
   localEventServer=,
   /*migrate=no,*/
   namevar=,
   numPeriods=*,
   offset=0,
   outputdir=,
   outputtype=f,
   outputdslib=,
   priority=10,
   project=,
   projectds=,
   projectid=,
   projectname=,
   pw=,
   ranges=,
   rangesds=,
   range=,
   rangeds=,
   resulttype=1,
   scorecard=,
   scorecardds=,
   scorecardid=,
   scorecardvar=,
   sentBy=,
   setup=,
   setupds=,
   softwarecomponent=Remote Services,
   sort=,
   synchronous=,
   templatename=,
   user=
   );
   %global eventserver appname version debug;
   %let appname = SPMBMF;
   %let version = 2.4HF2;
   %put NOTE: &appname &version;

   %local props prop_vals scorecardprop attributeprop elementprop cellprop cellformatprop
          setupprop test;
   %local projectprop rangesprop filterprop teststr delim outputdirslash setupoutputdirslash
          tmpFile diagramsprop;
   %local setupinputdirslash inputdirslash;

   /* validate the MIGRATE parameter
   %let migrate=%qupcase(&migrate);
   %if &migrate=1 or &migrate=YES or &migrate=Y %then %let migrate=YES;
   %else %let migrate=NO;
   */

   /* if supplied, initiallize the return code variable */
   %if %bquote(&bmfrcvar) ne %str( ) %then %let &bmfrcvar = 0;

   /* convert the action parameter values to uppercase and convert numeric values*/
   %let action=%qupcase(&action);
   %if &action=1 %then %let action=CREATE;
   %else %if &action=2 %then %let action=MODIFY;
   %else %if &action=3 %then %let action=GET;
   %else %if &action=%str() %then %let action=CREATE;

   /* must specify a user and pw */
   %if %bquote(&user) = %str() %then %do;
      %put ERROR: &appname &version - Cannot run &appname without specifying a user. ;
      %goto bmferror;
   %end;
   %if %bquote(&pw) = %str() %then %do;
      %put ERROR: &appname &version - Cannot run &appname without specifying a password.;
      %goto bmferror;
   %end;

   /* make project and template names safe for path names */
   %local alterredprojectname alterredtemplatename;
   %convertbaddirchars(inchars=&projectname,outvar=alterredprojectname,invalidchars=%bquote(&invalidDirChars));
   %convertbaddirchars(inchars=&templatename,outvar=alterredtemplatename,invalidchars=%bquote(&invalidDirChars));

   /* shape up the outputdir */
   %let outputdir=%bquote(&outputdir);
   %if %bquote(&outputdir) = %str() %then %do;
      %put ERROR: Cannot run &appname with a blank output dir. ;
      %goto bmferror;
   %end;
   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: &appname(DEBUG): outputdir = &outputdir;
   %end;
   %if %index(&outputdir,\) > 0 %then %do;
      %let delim = \;
   %end;
   %else %do;
      %let delim = /;
   %end;
   %if %index(%substr(&outputdir,%length(&outputdir),1),&delim) = 0 %then %do;
      %let outputdirslash = %trim(&outputdir)&delim;
   %end;
   %else %do;
      %let outputdirslash = &outputdir;
      %let outputdir = %substr(&outputdir,1,%eval(%length(&outputdir)-1));
   %end;

   /* add template name to the end of the outputdir since that is where the setup file */
   /* will be output to */
   %let setupoutputdirslash = %trim(&outputdirslash)%trim(&alterredtemplatename)&delim;
   /* add template name and project name to end of the outputdir since that is now the */
   /* root of the output from the midtier code. */
   %let outputdirslash = %trim(&setupoutputdirslash)%trim(&alterredprojectname)&delim;
   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: &appname(DEBUG): setupoutputdirslash = &setupoutputdirslash;
      %put %str( );
      %put NOTE: &appname(DEBUG): outputdirslash = &outputdirslash;
   %end;

   /* clean up the inputdir */
   %let inputdir=%nrbquote(&inputdir);
   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: &appname(DEBUG): inputdir = &inputdir;
   %end;
   %if %nrbquote(&inputdir) ne %str() %then %do;
      %if %index(&inputdir,\) > 0 %then %do;
         %let delim = \;
      %end;
      %else %do;
         %let delim = /;
      %end;
      %if %index(%substr(&inputdir,%length(&inputdir),1),&delim) = 0 %then %do;
         %let inputdirslash = %trim(&inputdir)&delim;
      %end;
      %else %do;
         %let inputdirslash = &inputdir;
         %let inputdir = %substr(&inputdir,1,%eval(%length(&inputdir)-1));
      %end;
      /* add template name to the end of the inputdir since that is where the setup */
      /* file will be input from */
      %let setupinputdirslash = %trim(&inputdirslash)%trim(&alterredtemplatename)&delim;
      /* add template name and project name to end of the inputdir since that is now */
      /* the root of the input from the midtier code. */
      %let inputdirslash = %trim(&setupinputdirslash)%trim(&alterredprojectname)&delim;
   %end;
   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: &appname(DEBUG): setupinputdirslash = &setupinputdirslash;
      %put %str( );
      %put NOTE: &appname(DEBUG): inputdirslash = &inputdirslash;
   %end;


   /* get the event server */

   /* set metauser and metapass options to the SPM user/pw if the metauser option */
   /* has not already been set */
   %let teststr = %sysfunc(getoption(metauser));
   %if %bquote(&teststr) = %str() %then %do;
      options metauser="&user" metapass="&pw";
   %end;

   /* if there is not a supplied eventserver then get one */
   %if &localEventServer = %str() %then %do;
      /* get a valid event listener to send the event to */
      %getlstnr(SoftwareComponentName=&softwarecomponent)
   %end;
   %else %do;
      /* if there is one, then assign it to a local copy of EVENTSERVER */
      %local eventserver;
      %let eventserver=&localEventServer;
   %end;

   /* if there is still no event server then abort */
   %if %bquote(&EVENTSERVER) = %str() %then %do;
      %put ERROR: Cannot run &appname without an event server. ;
      %goto bmferror;
   %end;
   %else %if %bquote(&debug) = %str(Y) %then %do;
      %put NOTE: &appname(DEBUG): EVENTSERVER = &EVENTSERVER;
   %end;

   %if %upcase(%bquote(&action))=CREATE or %bquote(&action)=MODIFY %then %do;
      /* If no input files or datasets are specified, default to files in the INPUTDIR directory if they are present. */
      /* If the input directory csv files are not there, and there is an INPUTDSLIB defined, and a dataset is  */
      /* there, then use the dataset as input. */
      %assignInput(&scorecardds, &scorecard, &inputdirslash&alterredprojectname, &alterredprojectname, &inputdslib, Scorecard)
      %assignInput(&elementds, &element, &inputdirslash&alterredprojectname, &alterredprojectname, &inputdslib, Element)
      %assignInput(&attributeds, &attribute, &inputdirslash&alterredprojectname, &alterredprojectname, &inputdslib, Attribute)
      %assignInput(&cellds, &cell, &inputdirslash&alterredprojectname, &alterredprojectname, &inputdslib, Cell)
      %assignInput(&projectds, &project, &inputdirslash&alterredprojectname, &alterredprojectname, &inputdslib, Project)

      %assignInput(&rangesds, &ranges, &inputdirslash&alterredprojectname, &alterredprojectname, &inputdslib, Range)
      /* this is an attempt to correct the fact that initially the parameter RANGES was used
      /* when it should have been RANGE. */
      %if %bquote(&range) ne %str() %then %let ranges = &range;
      %if %bquote(&rangeds) ne %str() %then %let rangesds = &rangeds;

      %assignInput(&filterds, &filter, &inputdirslash&alterredprojectname, &alterredprojectname, &inputdslib, Filter)
      %assignInput(&cellformatds, &cellformat, &inputdirslash&alterredprojectname, &alterredprojectname, &inputdslib, Cellformat)
      %assignInput(&setupds, &setup, &setupinputdirslash&alterredtemplatename, &alterredtemplatename, &inputdslib, Setup)
   %end;

   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: &appname(DEBUG): scorecard: &scorecard;
      %put NOTE: &appname(DEBUG): element = &element;
      %put NOTE: &appname(DEBUG): cell = &cell;
      %put NOTE: &appname(DEBUG): setup = &setup;
      %put NOTE: &appname(DEBUG): project = &project;
      %put NOTE: &appname(DEBUG): ranges = &ranges;
      %put NOTE: &appname(DEBUG): filter = &filter;
      %put NOTE: &appname(DEBUG): cellformat = &cellformat;
      %put NOTE: &appname(DEBUG): scorecardds = &scorecardds;
      %put NOTE: &appname(DEBUG): elementds = &elementds;
      %put NOTE: &appname(DEBUG): cellds = &cellds;
      %put NOTE: &appname(DEBUG): setupds = &setupds;
      %put NOTE: &appname(DEBUG): projectds = &projectds;
      %put NOTE: &appname(DEBUG): rangesds = &rangesds;
      %put NOTE: &appname(DEBUG): filterds = &filterds;
      %put NOTE: &appname(DEBUG): cellformatds = &cellformatds;
   %end;

   /* check for diagram input */
   %if %bquote(&diagrams) ne %str() %then %do;
      %let diagramsprop=diagrams;
      %if %bquote(&debug) = %str(Y) %then %do;
         %put NOTE: &appname(DEBUG): diagram input = &diagrams;
      %end;
   %end;

   /* set up the property variable for each type with the correct input file path */
   %setInputProperty(file=&scorecard,   sasds=&scorecardds,  propertyVariable=scorecardprop,  type=scorecard,  dummyrow=Y)
   %setInputProperty(file=&attribute,   sasds=&attributeds,  propertyVariable=attributeprop,  type=attribute,  dummyrow=Y)
   %setInputProperty(file=&element,     sasds=&elementds,    propertyVariable=elementprop,    type=element,    dummyrow=Y)
   %setInputProperty(file=&cellformat,  sasds=&cellformatds, propertyVariable=cellformatprop, type=cellformat, dummyrow=N)
   %setInputProperty(file=&cell,        sasds=&cellds,       propertyVariable=cellprop,       type=cell,       dummyrow=Y)
   %setInputProperty(file=&setup,       sasds=&setupds,      propertyVariable=setupprop,      type=setup,      dummyrow=N)
   %setInputProperty(file=&ranges,      sasds=&rangesds,     propertyVariable=rangesprop,     type=ranges,     dummyrow=N)
   %setInputProperty(file=&filter,      sasds=&filterds,     propertyVariable=filterprop,     type=filterfile, dummyrow=N)
   %setInputProperty(file=&project,     sasds=&projectds,    propertyVariable=projectprop,    type=project,    dummyrow=N)

   /* set the caller name */
   %if %bquote(&sentby) = %str() %then %let sentby=&appname.Caller;

   /* set the eventname and synchronous flag */
   %local defaultEventName defaultSyncEventName;
   %let defaultEventName =SAS.Solutions.SpmBOBInterface;
   %let defaultSyncEventName =SAS.Solutions.SpmBOBInterfaceSynch;

   %if %bquote(&synchronous)= Y or %bquote(&synchronous) = YES or %bquote(&synchronous) = 1 %then %do;
      %if %bquote(&eventname)=%str() %then %do;
         %let eventName=&defaultSyncEventName;
         %let synchronous=Y;
      %end;
      %else %if %bquote(&eventname)= &defaultEventName %then
         %let synchronous = N;
   %end;
   %else %do;
      %if %bquote(&eventname)=%str() %then %do;
         %let eventName=&defaultEventName;
         %let synchronous = N;
      %end;
      %else %if %bquote(&eventName) = &defaultSyncEventName %then %do;
         %let synchronous = Y;
      %end;
      %else %let synchronous = N;
   %end;
   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: &appname(DEBUG): synchronous = &synchronous;
      %put NOTE: &appname(DEBUG): eventName = &eventName;
   %end;

   /* construct the event properties */
   %let sendevnt_delimiter=%str(^);

   /* construct the event properties */
   %let props=outputdir outputtype bobuser sentby pw entitykey action resulttype /*migrate*/;
   %let prop_vals=&outputdir&sendevnt_delimiter&outputtype&sendevnt_delimiter;
   %let prop_vals=&prop_vals&user&sendevnt_delimiter&sentby&sendevnt_delimiter&pw;
   %let prop_vals=&prop_vals&sendevnt_delimiter&entitykey&sendevnt_delimiter&action;
   %let prop_vals=&prop_vals&sendevnt_delimiter&resulttype/*&sendevnt_delimiter&migrate*/;

   /* add setup csv file if present. */
   %addpropname(propname=%nrbquote(&setupprop),propval=%nrbquote(&setup))
   /* add project csv file if present. */
   %addpropname(propname=%nrbquote(&projectprop),propval=%nrbquote(&project))
   /* add range csv file if present. */
   %addpropname(propname=%nrbquote(&rangesprop),propval=%nrbquote(&ranges))
   /* add diagram file if present. */
   %addpropname(propname=%nrbquote(&diagramsprop),propval=%nrbquote(&diagrams))
   /* add scorecard csv file if present. */
   %addpropname(propname=%nrbquote(&scorecardprop),propval=%nrbquote(&scorecard))
   /* add element csv file if present. */
   %addpropname(propname=%nrbquote(&elementprop),propval=%nrbquote(&element))
   /* add attribute csv file if present. */
   %addpropname(propname=%nrbquote(&attributeprop),propval=%nrbquote(&attribute))
   /* add cell csv file if present. */
   %addpropname(propname=%nrbquote(&cellprop),propval=%nrbquote(&cell))
   /* add cell format csv file if present. */
   %addpropname(propname=%nrbquote(&cellformatprop),propval=%nrbquote(&cellformat))
   /* add filter csv file if present. */
   %addpropname(propname=%nrbquote(&filterprop),propval=%nrbquote(&filter))
   /* add domain if present. */
   %addpropname(propname=bobdomain,propval=%nrbquote(&domain),valcheck=Y)
   /* add projectid if present. */
   %addpropname(propname=projectid,propval=%nrbquote(&projectid),valcheck=Y)
   /* add scorecardid if present. */
   %addpropname(propname=scorecardid,propval=%nrbquote(&scorecardid),valcheck=Y)
   /* add elementtypeid if present. */
   %addpropname(propname=elementtypeid,propval=%nrbquote(&elementtypeid),valcheck=Y)
   /* add date if present. */
   %addpropname(propname=date,propval=%nrbquote(&date),valcheck=Y)
   /* add offset if present. */
   %addpropname(propname=offset,propval=%nrbquote(&offset),valcheck=Y)
   /* add numPeriods if present. */
   %addpropname(propname=numPeriods,propval=%nrbquote(&numPeriods),valcheck=Y)
   /* add sort if present. */
   %addpropname(propname=sort,propval=%nrbquote(&sort),valcheck=Y)
   /* add filterid if present. */
   %addpropname(propname=filterid,propval=%nrbquote(&filterid),valcheck=Y)
   /* add cultureid if present. */
   %addpropname(propname=cultureid,propval=%nrbquote(&cultureid),valcheck=Y)
   /* add projectname if present. */
   %addpropname(propname=projectname,propval=%nrbquote(&projectname),valcheck=Y)
   /* add templatename if present. */
   %addpropname(propname=templatename,propval=%nrbquote(&templatename),valcheck=Y)
   /* add namevar if present. */
   %addpropname(propname=namevar,propval=%nrbquote(&namevar),valcheck=Y)
   /* add scorecardvar if present. */
   %addpropname(propname=scorecardvar,propval=%nrbquote(&scorecardvar),valcheck=Y)
   /* add badnumber if present. */
   %addpropname(propname=badnumber,propval=%nrbquote(&badnumber),valcheck=Y)
   /* add invalid directory characters if present. */
   %addpropname(propname=invalidDirChars,propval=%nrbquote(&invalidDirChars),valcheck=Y)


   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: &appname(DEBUG): props = &props;
      %put %str( );
      %put NOTE: &appname(DEBUG): prop_vals = &prop_vals;
   %end;

   /* run the event either synchronously or asynchronously */
   %local result_url result_rc;
   %let result_rc=0;
   %if &synchronous=Y %then %do;
      filename foo temp;
      %let result_url = %bquote(%sysfunc(pathname(foo)));
      %if %bquote(&debug) = %str(Y) %then %do;
         %put %str( );
         %put NOTE: &appname(DEBUG): Sending synchronous event;
      %end;

      %sendevnt(evntname=&eventName,
         priority=&priority,
         eventserver=&EVENTSERVER,
         sentby=&sentby,
         response=RESULT,
         result_url=&result_url,
         vals_delimiters=&sendevnt_delimiter,
         props=%bquote(&props),
         vals=%bquote(&prop_vals)
      )
   %end;
   %else %do;
      %if %bquote(&debug) = %str(Y) %then %do;
         %put %str( );
         %put NOTE: &appname(DEBUG): Sending asynchronous event;
      %end;
      %sendevnt(evntname=&eventName,
         priority=&priority,
         eventserver=&EVENTSERVER,
         sentby=&sentby,
         vals_delimiters=&sendevnt_delimiter,
         props=%bquote(&props),
         vals=%bquote(&prop_vals)
      )

   %end;

   /* check to see if csv output needs to be converted to datasets */
   %if &synchronous = Y and %bquote(&outputdslib) ne %str() %then %do;
      %if %sysfunc(libref(%bquote(&outputdslib))) ne 0 %then %do;
         %put WARNING: %upcase(%bquote(&outputdslib)) is not assigned, cannot create output SAS data sets.;
      %end;
      %else %do;
         %procimport(csvpath=&outputdirslash&alterredprojectname._Scorecard.csv,sasds=&outputdslib..&alterredprojectname._Scorecard,
               guessingrows=&guessingrows)
         %procimport(csvpath=&outputdirslash&alterredprojectname._Element.csv,sasds=&outputdslib..&alterredprojectname._Element,
              guessingrows=&guessingrows)
         %procimport(csvpath=&outputdirslash&alterredprojectname._Attribute.csv,sasds=&outputdslib..&alterredprojectname._Attribute,
               guessingrows=&guessingrows)
         %procimport(csvpath=&outputdirslash&alterredprojectname._Cell.csv,sasds=&outputdslib..&alterredprojectname._Cell,
               guessingrows=&guessingrows)
         %procimport(csvpath=&outputdirslash&alterredprojectname._CellFormat.csv,sasds=&outputdslib..&alterredprojectname._CellFormat,
               guessingrows=&guessingrows)
         %procimport(csvpath=&outputdirslash&alterredprojectname._Range.csv,sasds=&outputdslib..&alterredprojectname._Range,
               guessingrows=&guessingrows)
         %procimport(csvpath=&setupoutputdirslash&alterredtemplatename._Setup.csv,sasds=&outputdslib..&alterredtemplatename._Setup,
               guessingrows=&guessingrows)
         %procimport(csvpath=&outputdirslash&alterredprojectname._Project.csv,sasds=&outputdslib..&alterredprojectname._Project,
               guessingrows=&guessingrows)
      %end;
   %end;

   %if &synchronous ne Y %then %do;
      %put NOTE: &appname &version has ended but some &appname processes may still be running asynchronously.;
   %end;
   %else %do;
      %put NOTE: &appname &version ended successfully;
   %end;
   %goto exit;

   %bmferror:
   %put NOTE: &appname &version ended with errors.;
   %if &bmfrcvar ne %str() %then %do;
      %if &result_rc gt 0 %then %let &bmfrcvar = &result_rc;
      %else %let &bmfrcvar=1;
   %end;
   %exit:

%mend;

%macro addpropname(propname=,propval=,valcheck=);
   /* Adds &propname to the props string.  Adds &propval to the prop_vals string. */
   /* valcheck= controls whether the operation is based on a non blank value or non-blank property name */
   %if (%nrbquote(&valcheck) eq Y and %nrbquote(&propval) ne %str()) or
      (%nrbquote(&valcheck) ne Y and %nrbquote(&propname) ne %str()) %then %do;
      %let props=&props %nrbquote(&propname);
      %let prop_vals=&prop_vals&sendevnt_delimiter%nrbquote(&propval);
   %end;
%mend;

%macro assignInput(dsvar, nondsvar, fileprefix, libprefix, inputdslib, type);
   /* assign a file to &type if there is not one already specified */

   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: assignInput(DEBUG): dsvar=&dsvar, nondsvar=&nondsvar, fileprefix=&fileprefix, libprefix=&libprefix, inputdslib=&inputdslib, type=&type;
   %end;
   /* if there is no file or SAS DS specification then check if there is a file in the inputdirectory */
   %if %bquote(&dsvar) = %str() and %bquote(&nondsvar) = %str() %then %do;
      %if %sysfunc(fileexist(&fileprefix._&type..csv)) %then %do;
         %let &type=&fileprefix._&type..csv;
         %if %bquote(&debug) = %str(Y) %then %do;
            %put NOTE: assignInput(DEBUG): &fileprefix._&type..csv: fileexist = %sysfunc(fileexist(&fileprefix._&type..csv));
         %end;
      %end;
      /* then check if a SAS data set exists in the supplied libref */
      %else %if %sysfunc(exist(&inputdslib..&libprefix._&type))=1 %then %do;
          %let &type.ds=&inputdslib..&libprefix._&type;
          %if %bquote(&debug) = %str(Y) %then %do;
             %put NOTE: assignInput(DEBUG): &inputdslib..&libprefix._&type: exist = %sysfunc(exist(&inputdslib..&libprefix._&type));
          %end;
      %end;
   %end;

%mend;

%macro procimport(csvpath=,sasds=,guessingrows=32767);
   /* converts a csv file name in &csvpath to a sas data set name in &sasds */
   %if %bquote(&debug) = %str(Y) %then %do;
      %put NOTE: &appname(DEBUG): Converting csv file to SAS datasets.;
      %put NOTE: &appname(DEBUG): csv: &csvpath;
      %put NOTE: &appname(DEBUG): sasds: &sasds;
   %end;

   %if %sysfunc(fileexist(&csvpath)) = 0 %then %do;
      %if %bquote(&debug) = %str(Y) %then %do;
         %put WARNING: &appname(DEBUG): &csvpath does not exist.;
      %end;
   %end;

   %else %do;
      /* source is being turned off because PROC IMPORT produces a lot of generated source */
      %if %bquote(&debug) ne Y %then %do;
         options nosource;
      %end;
      proc import DATAFILE="&csvpath" OUT=&sasds DBMS=CSV REPLACE;
         guessingrows=&guessingrows;
      run;
      %if %bquote(&debug) ne Y %then %do;
         options source;
      %end;
   %end;
%mend;

%macro setInputProperty(file=, sasds=, propertyVariable=, type=, dummyRow=Y);
   /* set the inputfile for the input type.  The file can be either a CSV file or SAS ds.  */
   /* If it is a SAS ds then it will be converted to a CSV file */
   /* file = the name of the input CSV file for this type. */
   /* sasds = the name of the SAS ds.  This will be converted to CSV file */
   /* propertyVariable = the name of the macro variable that will hold the path to the CSV file */
   /* type = the input file type, i.e.  scorecard or cell or setup or element, etc. */

   /* if there is no specified CSV file */
   %if %bquote(&file) = %str() %then %do;
      /* and if there is no specified SAS data set, then return */
      %if %bquote(&sasds) = %str() %then %do;
         /* no data set either */
         %let &propertyVariable=;
         %if %bquote(&debug) = %str(Y) %then %do;
            %put NOTE: &appname(DEBUG): No &type input;
         %end;
      %end;
      /* if there is a specified SAS ds */
      %else %do;
         /* set up a macro variable by the same name as the type.  This will be used */
         /* later to add the property to the list to be published */
         %let &propertyVariable=&type;
         /* convert the SAS ds to a CSV file */
         %createCSVInputFile(ds=&sasds, f=tmpFile, path=&converteddsdir,dummyRow=&dummyRow)
         %let &type=&tmpFile;
         %if %bquote(&debug) = %str(Y) %then %do;
            %put NOTE: &appname(DEBUG): &type input from SAS dataset written to csv: &&&type;
         %end;
      %end;
   %end;
   /* there is a specified CSV file, so use it */
   %else %do;
      %let &propertyVariable=&type;
      %if %bquote(&debug) = %str(Y) %then %do;
         %put NOTE: &appname(DEBUG): &type input: &file;
      %end;
   %end;
%mend;


%macro createCSVInputFile(ds=, f=, path=, dummyRow=N);
   /* Creates a CSV file based on an input SAS ds and returns the complete path    */
   /*  and filename of the CSV file                                                */
   /*     ds = name of the SAS data set to be used as the basis for conversion     */
   /*     f = the name of the macro variable that will hold the completed path     */
   /*     path = the directory where the CSV file will be created                  */
   /*     dummyRow = used only by the IMAP facility, not documented here           */
   %if %bquote(&debug) = %str(Y) %then %do;
      %put %str( );
      %put NOTE: &appname(DEBUG): Converting SAS dataset %upcase(&ds) to CSV file %upcase(&ds).CSV.;
      %if %bquote(&path) = %str() %then %put NOTE: &appname(DEBUG): No path specified, default will be used.;
      %else %put NOTE: &appname(DEBUG): CSV path is &path..;
   %end;

   %local i word source notes optname fid slash;
   %if &sysscp=WIN %then %let slash=\;
   %else %let slash=/;
   %if %bquote(&path) = %str() %then %do;
      /* use a default path. This will only work on a single tier setup */
      %if &sysscp=WIN %then
      %let &f = !SASROOT/scorecard/sasmisc/&ds..csv;
      %else
      %let &f = !SASROOT/misc/scorecard/&ds..csv;
      %if %bquote(&debug) = %str(Y) %then %do;
         %put WARNING: &appname(DEBUG): Default CSV file location of &&&f will be used.  Verify that this is a valid path for the BMF server.;
      %end;
   %end;
   /* otherwise combine the path and the csv file name.  The csv file will be the sas ds name */
   /* plus the .csv extension.  */
   %else %do;
      /* if necessary, add a slash at the end of the path */
      %let test = %substr(&path, %length(&path), 1);
      %if %bquote(&test) = %str(/) or %bquote(&test) = %str(\) %then %do;
         %let &f=&path.&ds..csv;
      %end;
      %else %do;
         %let &f=&path&slash&ds..csv;
      %end;
   %end;

   /* begin the conversion */

   /* get full, resolved path name */
   filename tmpfile "&&&f";
   %let fid=%sysfunc(fopen(tmpfile));
   %if &fid gt 0 %then %do;
      %let optname=%sysfunc(foptname(&fid, 1));
      %let &f=%sysfunc(finfo(&fid,&optname));
      %let fid=%sysfunc(fclose(&fid));
   %end;

   /* if the file exists and cannot be deleted then abort the whole BMF call */
   %if %sysfunc(fileexist(&&&f)) = 1 %then %do;
      %let rc = %sysfunc(fdelete(tmpfile));
      %if &rc ne 0 %then %do;
         %put ERROR: Attempting to convert SAS dataset %upcase(&ds) to CSV file but &&&f previously exists and cannot be replaced.;
         %let &f=;
         %goto exit;
      %end;
      %if %bquote(&debug) = %str(Y) %then %do;
         %put NOTE: &appname(DEBUG): file to write out to is &&&f;
      %end;
   %end;

   options nonotes;
   filename tmpfile clear;
   options notes;

   %getvars(dsname=&ds);

   options notes missing=' ';
   %let rc=0;
   data _null_;
      set &ds end=end;
      file "&&&f" LINESIZE=32767;

      /* put out a dummy record at the top since java will discard it. */
      dummyRow = "&dummyRow";
      if _N_ = 1 & dummyRow="Y" then do;
         put "please ignore this row, BMF will consume it later.";
      end;

      /* generate appropriate put statement for each variable */
      %let i = 1;
      %do %until (&word=%str());
         %let word=%scan(&vars, &i, %str( ));
         %if &word ne %str() %then %do;
            if index(&word, ',') > 0 then
            put '"' &word +(-1)'",' @;
            else if &word = '' then
            put ',' @;
            else
            put &word  +(-1)',' @;
         %end;
         %let i = %eval(&i + 1);
      %end;
      put +(-1)' ';
      if end then call symput('RC','1');
   run;
   %if &rc = 0 %then %do;
      %put ERROR: BMF is not able to convert SAS data set &ds to &&&f;
   %end;
   %if %bquote(&debug) = %str(Y) %then %do;
      %put NOTE: Successfully created &&&f..;
   %end;
   %exit:

   options missing='.';

%mend;

%macro getvars(dsname=,pw=);
   %local i n;
   %global vars;
   %let n=0;
   %let vars=;
   %if &pw ne %then %let pw=read=&pw;
   options nodsnferr nonotes nosource;
   proc contents varnum data=&dsname(&pw) noprint out=_varlst_(keep=name varnum);
   run;
   proc sort data=_varlst_; by varnum;
   run;
   data _null_;
      set _varlst_ end=end;
      call symput('n'!!left(put(_n_,3.)),name);
      if end then call symput('n',left(put(_n_,3.)));
   run;
   proc datasets lib=work nolist;delete _varlst_;quit;
   run;
   options &notes &source;
   %do i=1 %to &n;
      %let vars=%trim(&vars) &&n&i;
   %end;
   %let vars=%str( &vars );
   options notes source;
%mend;

%macro convertbaddirchars(inchars=,outvar=,invalidChars=);
     /* converts any invalid characters in the INCHARS parameter and stores the result */
     /* in the variables defined in the OUTVAR parameter.  Invalid characters are  */
     /* defined as those in the INVALIDCHARS parameter. */
     %if &outvar ne %str() %then %do;
       %if %nrbquote(&invalidChars) = %str() %then
         %let invalidChars=%str( !@#$%^&*()+-={}|[]\:<>?,./''"");
       %local convert;
       /* the convert string must be as leaset as long as the invalid chars string */
       %let convert=%sysfunc(repeat(_,%length(&invalidChars)));

       %if %length(%nrbquote(&inchars)) > 0 %then %do;
          %let &outvar = %sysfunc(translate(%nrbquote(&inchars), &convert, &invalidChars));
       %end;
       %else %do;
         %let &outvar = %str();
       %end;
     %end;
%mend;

/*
Sample usage.
Replace values in <...> with appropriate values for your site.  The arguments can be
in any order


options metaserver=<your-metadata-server> metaport=8561 metarepository=Foundation;
%let debug=Y;
%spmbmf(action=MODIFY,
   user=<user-id>,
   pw=<password>,
   templatename=<your-template-name>,
   projectname=<your-project-name>,

   **** after this can be some number of input file arguments ****
   scorecard=<scorecard-file.csv>,
   attribute=<attribute-file.csv>,

   outputdir=<output-folder>
   );
   */

   /* run this to get your encrypted password */
   /* proc pwencode in='pw_to_encrypt'; run; */

   /* sample data */
   /*
   data attribute;
   input @1 var1 $1. @3 var2 $12. @16 var3 $9. @26 var4 $21.;
   cards;
   1 text         MyText    MyLabel
   1 email        MyEmail   myEmailAddress@sas.com
   1 date         MyDate    1/24/2006
   1 link         MyLink    www.yahoo.com
   1 Element Type MyElement 3
   ;
   run;
   */
